/***************************************************************************
                             graphwidget.cpp
                             -------------------
    begin                : Fri Jun 10 2005
    copyright            : (C) 2005 by Elad Lahav
    email                : elad_lahav@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qpainter.h>
#include <qtooltip.h>
#include <klocale.h>
#include "graphwidget.h"
#include "codegenerator.h"
#include "graphnode.h"
#include "graphedge.h"
#include "kscopeconfig.h"

#define GRAPH_WIDGET_NAME	(char*)"graphwidget"

#define SET_GRAPH_ATTR(_attr, _val) \
	agxset(m_pGraph, agraphattr(m_pGraph, (char*)_attr, (char*)"")->index, \
		(char*)_val)
#define GET_GRAPH_ATTR(_attr) \
	agxget(m_pGraph, agraphattr(m_pGraph, (char*)_attr, (char*)"")->index)

#define SET_NODE_ATTR(_node, _attr, _val) \
	agxset(_node, agnodeattr(m_pGraph, (char*)_attr, (char*)"")->index, \
		(char*)_val)
#define GET_NODE_ATTR(_node, _attr) \
	agxget(_node, agnodeattr(m_pGraph, (char*)_attr, (char*)"")->index)

#define SET_EDGE_ATTR(_edge, _attr, _val) \
	agxset(_edge, agedgeattr(m_pGraph, (char*)_attr, (char*)"")->index, \
		(char*)_val)
#define GET_EDGE_ATTR(_edge, _attr) \
	agxget(_edge, agedgeattr(m_pGraph, (char*)_attr, (char*)"")->index)

#define FILE_VERSION		3

typedef Agedge_t* (*FirstEdgeFunc)(Agraph_t*, Agnode_t*);
typedef Agedge_t* (*NextEdgeFunc)(Agraph_t*, Agedge_t*);

char ATTR_TRUE[] = "t";
char ATTR_FALSE[] = "f";

const char* GRAPH_DIRS[] = { "TB", "LR", "BT", "RL" };

/**
 * Displays a tool tip on the graph.
 * Note that we cannot use the standard tool tip class here, since graph
 * items are neither rectangular nor is their position known in advance.
 * @author	Elad Lahav
 */
class GraphTip : public QToolTip
{
public:
	/**
	 * Class constructor.
	 * @param	pWidget	Owner graph widget
	 */
	GraphTip(GraphWidget* pWidget) : QToolTip(pWidget->viewport()),
		m_pGraphWidget(pWidget) {}
	
	/**
	 * Class destructor.
	 */
	virtual ~GraphTip() {}
	
protected:
	/**
	 * Called when the pre-conditions for a tool tip are met.
	 * Asks the owner for a tip to display and, if one is returned, shows
	 * the tool tip.
	 * @param	ptPos	Current mouse position
	 */
	virtual void maybeTip(const QPoint& ptPos) {
		QString sText;
		QRect rc;
		
		// Display a tip, if required by the owner
		sText = m_pGraphWidget->getTip(ptPos, rc);
		if (sText != QString::null)
			tip(rc, sText);
	}
	
private:
	GraphWidget* m_pGraphWidget;
};

GVC_t* GraphWidget::s_pGVC;

/**
 * Class constructor.
 * @param	pParent	The parent widget
 * @param	szName	The widget's name
 */
GraphWidget::GraphWidget(QWidget* pParent, const char* szName) :
	QCanvasView(pParent, szName),
	m_pCanvas(NULL),
	m_pGraph(NULL),
	m_pen(QColor(0, 0, 0)),
	m_brush(QColor(255, 255, 255)),
	m_font(),
	m_pCurItem(NULL),
	m_dZoom(1.0)
{
	// Create the popup menus
	m_pNodePopup = new QPopupMenu(this);
	m_pNodePopup->insertItem(i18n("Show Called Functions"), this,
		SLOT(slotShowCalled()));
	m_pNodePopup->insertItem(i18n("Show Calling Functions"), this,
		SLOT(slotShowCalling()));
	m_pNodePopup->insertSeparator();
	m_pNodePopup->insertItem(i18n("Hide Called Functions"), this,
		SLOT(slotHideCalled()));
	m_pNodePopup->insertItem(i18n("Hide Calling Functions"), this,
		SLOT(slotHideCalling()));
	m_pNodePopup->insertSeparator();
	m_pNodePopup->insertItem(i18n("Find Definition"), this,
		SLOT(slotFindDef()));
	
	(void)new GraphTip(this);
	
	m_pEdgePopup = new QPopupMenu(this);
	m_pEdgePopup->insertItem(i18n("Open Call"), this, SLOT(slotOpenCall()));
	
	// Create a new graph
	reset();
}

/**
 * Class destructor.
 */
GraphWidget::~GraphWidget()
{
	if (m_pGraph != NULL)
		agclose(m_pGraph);
		
	if (m_pCanvas != NULL)
		delete m_pCanvas;
}

/**
 * Resizes the canvas.
 * @param	size	The new size to use
 */
void GraphWidget::resize(const QSize& size)
{
	m_pCanvas->resize(size.width(), size.height());
}

/**
 * Creates a new graphviz graph object.
 * If there was a previous graph, it is deleted.
 */
void GraphWidget::reset()
{
	// Clean up
	if (m_pGraph != NULL)
		agclose(m_pGraph);
		
	// Create a new graph, and set its attributes
	m_pGraph = agopen(GRAPH_WIDGET_NAME, AGDIGRAPHSTRICT);
	SET_GRAPH_ATTR("bgcolor", 
		Config().getColor(KScopeConfig::GraphBack).name().latin1());
	SET_GRAPH_ATTR("rankdir", Config().getGraphOrientation().latin1());
	
	// Set default node properties
	setNodeAttrib();
}

/**
 * Creates a root node for the graph.
 * The root node defines the connected component which is always displayed
 * (all other connected components are removed when they are no longer
 * strongly connected to the root).
 * @param	sFunc	The function name for the root node
 */
void GraphWidget::setRoot(const QString& sFunc)
{
	// Insert a new node to the graph
	(void)agnode(m_pGraph, (char*)sFunc.latin1());
}

/**
 * Adds a call to the graph.
 * A call is made between two functions, the caller and the callee.
 * @param	data	Contains information on the call
 */
void GraphWidget::addCall(const CallData& data)
{
	Agnode_t* pCaller, * pCallee;
	Agedge_t* pEdge;
	
	// Find the relevant nodes (create new nodes if necessary)
	pCaller = agnode(m_pGraph, (char*)data.m_sCaller.latin1());
	pCallee = agnode(m_pGraph, (char*)data.m_sCallee.latin1());
	
	// Create a directed edge
	pEdge = agedge(m_pGraph, pCaller, pCallee);
	SET_EDGE_ATTR(pEdge, "kscope_file", data.m_sFile.latin1());
	SET_EDGE_ATTR(pEdge, "kscope_line", data.m_sLine.latin1());
	SET_EDGE_ATTR(pEdge, "kscope_text", data.m_sText.latin1());
}

/**
 * Draws the graph on the canvas using the graphviz engine.
 * A new canvas is created, so all items need to be regenerated.
 * TODO: Can we use the same canvas and only reposition existing items?
 */
void GraphWidget::draw()
{
	QWMatrix mtx;
		
	// Delete the old canvas
	if (m_pCanvas != NULL)
		delete m_pCanvas;
		
	// Create a new canvas
	m_pCanvas = new QCanvas;
	setCanvas(m_pCanvas);
	
	// Apply the zoom factor
	mtx.scale(m_dZoom, m_dZoom);
	setWorldMatrix(mtx);

	// Attach to the code generator
	CodeGenerator::bind(this);
	gvBindContext(s_pGVC, m_pGraph);
	
	// Lay out and draw the graph
	dot_layout(m_pGraph);
 	emit_graph(s_pGVC, EMIT_SORTED);
	dot_cleanup(m_pGraph);
}

/**
 * Loads a graph from a file.
 * The file should contain a graph represented in the dot language.
 * @param	sFile	The file's path
 * @return	true if successful, false otherwise
 */
bool GraphWidget::load(const QString& sFile)
{
	FILE* pFile;
	int nVersion;
	
	// Open the file
	pFile = fopen(sFile.latin1(), "r");
	if (pFile == NULL)
		return false;
	
	// Verify the file version
	if ((fscanf(pFile, "VERSION=%d\n", &nVersion) != 1) || 
		(nVersion != FILE_VERSION)) {
		fclose(pFile);
		return false;
	}
	
	// Read the zoom factor
	m_dZoom = 1.0;
	fscanf(pFile, "Zoom=%lf\n", &m_dZoom);
	
	// Read the graph stored on this file
	m_pGraph = agread(pFile);
	fclose(pFile);
	
	if (m_pGraph == NULL)
		return false;
		
	// Set default node attributes
	setNodeAttrib();
	return true;
}

/**
 * Stores a graph on a file.
 * The file uses the dot language to describe the graph.
 * @param	sFile	The full path of the file to use
 * @param	bHeader	true to write a file header, false for a pure dot file
 */
void GraphWidget::save(const QString& sFile, bool bHeader)
{
	FILE* pFile;
	
	// Open/create the file
	pFile = fopen(sFile.latin1(), "w+");
	if (pFile == NULL)
		return;
		
	// Write the header, if requested
	if (bHeader) {
		fprintf(pFile, "VERSION=%d\n", FILE_VERSION);
		fprintf(pFile, "Zoom=%lf\n", m_dZoom);
	}
	
	// Write the graph using the dot language
	gvBindContext(s_pGVC, m_pGraph);
	dot_layout(m_pGraph);
	agwrite(m_pGraph, pFile);
	dot_cleanup(m_pGraph);
	
	// Clean up
	fclose(pFile);
}

/**
 * Changes the zoom factor.
 * @param	bIn	true to zoom in, false to zoom out
 */
void GraphWidget::zoom(bool bIn)
{
	QWMatrix mtx;
	
	// Set the new zoom factor
	if (bIn)
		m_dZoom *= 2.0;
	else
		m_dZoom /= 2.0;
		
	// Apply the transformation matrix
	mtx.scale(m_dZoom, m_dZoom);
	setWorldMatrix(mtx);
}

/**
 * Changes the graph's direction 90 degrees counter-clockwise.
 */
void GraphWidget::rotate()
{
	QString sDir;
	int i;
	
	// Get the current direction
	sDir = GET_GRAPH_ATTR("rankdir");
	
	// Find the next direction
	for (i = 0; i < 4 && sDir != GRAPH_DIRS[i]; i++);
	if (i == 4)
		i = 0;
	else
		i = (i + 1) % 4;

	// Set the new direction
	sDir = GRAPH_DIRS[i % 4];
	SET_GRAPH_ATTR("rankdir", sDir.latin1());
	Config().setGraphOrientation(sDir);
}

/**
 * Checks if a tool tip is required for the given position.
 * NOTE: We currently return a tool tip for edges only
 * @param	ptPos	The position to query
 * @param	rc		Holds the tip's rectangle, upon return
 * @return	The tip's text, or QString::null if no tip is required
 */
QString GraphWidget::getTip(const QPoint& ptPos, QRect& rc)
{
	QPoint ptRealPos, ptTopLeft, ptBottomRight;
	QCanvasItemList il;
	QCanvasItemList::Iterator itr;
	GraphEdge* pEdge;
	QString sText, sFile, sLine;
	
	ptRealPos = viewportToContents(ptPos);
	ptRealPos /= m_dZoom;
	pEdge = NULL;
	
	// Check if there is an edge at this position
	il = m_pCanvas->collisions(ptRealPos);
	for (itr = il.begin(); itr != il.end(); ++itr) {
		pEdge = dynamic_cast<GraphEdge*>(*itr);
		if (pEdge != NULL)
			break;
	}
	
	// No tip if no edge was found
	if (pEdge == NULL)
		return QString::null;
	
	// Set the rectangle for the tip (the tip is closed when the mouse leaves
	// this area)
	rc = pEdge->boundingRect();
	ptTopLeft = rc.topLeft();
	ptBottomRight = rc.bottomRight();
	ptTopLeft *= m_dZoom;	
	ptBottomRight *= m_dZoom;
	ptTopLeft = contentsToViewport(ptTopLeft);
	ptBottomRight = contentsToViewport(ptBottomRight);
	rc = QRect(ptTopLeft, ptBottomRight);
	
	// Create a tip for this edge
	sText = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_text");
	sFile = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_file");
	sLine = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_line");
			
	sText += "<br><i>" + sFile + "</i>:" + sLine;
	return sText;
}

/**
 * Initialises the graphviz engine and the graph drawing mechanism.
 * This function _must_ be called before any GraphWidget object is used.
 */
void GraphWidget::init()
{
	char* pArgs[] = {};
	
	s_pGVC = gvContext();
	dotneato_initialize(s_pGVC, 0, pArgs);
}

/**
 * Frees resources allocated to the graphviz engine.
 * Call this function when this class is no longer required.
 */
void GraphWidget::fini()
{
	dotneato_terminate(s_pGVC);
}

/**
 * Called before a node is drawn.
 * Creates a new GraphNode item.
 * @param	sName	The name of the node
 */
void GraphWidget::startNode(Agnode_t* pNode)
{
	m_pCurItem = new GraphNode(m_pCanvas, pNode);
}

/**
 * Called after a node has been drawn.
 * Displays the newly created GraphNode item.
 */
void GraphWidget::endNode()
{
	m_pCurItem->show();
	m_pCurItem = NULL;
}

/**
 * Called before an edge is drawn.
 * Creates a new ArrowEdge item.
 */
void GraphWidget::startEdge(Agedge_t* pEdge)
{
	m_pCurItem = new GraphEdge(m_pCanvas, pEdge);
}

/**
 * Called after an edge has been drawn.
 * Displays the newly created ArrowEdge item.
 */
void GraphWidget::endEdge()
{
	m_pCurItem->show();
	m_pCurItem = NULL;
}

/**
 * Draws a polygon using the given points.
 * A polygon is drawn for either a GraphNode, or for the arrow of an
 * ArrowEdge.
 * @param	arrPoints	Points outlining the polygon
 */
void GraphWidget::polygon(const QPointArray& arrPoints)
{
	GraphNode* pNode;
	GraphEdge* pEdge;
	
	// A NULL item means that graphviz is painting the graph's background.
	// We simply set the background colour of the canvas, based on the current
	// brush.
	if (m_pCurItem == NULL) {
		m_pCanvas->setBackgroundColor(m_brush.color());
		return;
	}
	
	// Set the polygon's fill and edge colours
	m_pCurItem->setBrush(m_brush);
	m_pCurItem->setPen(m_pen);
	
	// Set the polygon's coordinates
	if ((pNode = dynamic_cast<GraphNode*>(m_pCurItem)) != NULL)
		pNode->setPoints(arrPoints);
	else if ((pEdge = dynamic_cast<GraphEdge*>(m_pCurItem)) != NULL)
		pEdge->setPolyPoints(arrPoints);
}

/**
 * Draws the given text at the given position.
 * Currently text is only drawn inside a GraphNode canvas item.
 * @param	sText	The text to draw
 * @param	ptPos	The position of the text
 */
void GraphWidget::text(const QString& sText, const QPoint& ptPos)
{
	GraphNode* pNode;
	
	if ((pNode = dynamic_cast<GraphNode*>(m_pCurItem)) != NULL) {
		pNode->setFont(m_font);
		pNode->setText(sText);
		pNode->setTextPos(ptPos);
	}
}

/**
 * Draws a Bezier curve using the given control points.
 * @param	arrPoints	The control points to use
 */
void GraphWidget::curve(const QPointArray& arrPoints)
{
	GraphEdge* pEdge;
	
	pEdge = dynamic_cast<GraphEdge*>(m_pCurItem);
	if (pEdge != NULL) {
		pEdge->setPen(m_pen);
		pEdge->setCurvePoints(arrPoints);
	}
}

/**
 * Handles mouse clicks over the graph view.
 * @param	pEvent	Includes information on the mouse press event
 */
void GraphWidget::contentsMousePressEvent(QMouseEvent* pEvent)
{
	QPoint ptRealPos;
	QCanvasItemList il;
	QCanvasItemList::Iterator itr;
	QString sFunc;
	GraphNode* pNode;
	GraphEdge* pEdge;
	
	pNode = NULL;
	pEdge = NULL;
	
	// Handle right-clicks only
	if (pEvent->button() != Qt::RightButton) {
		QCanvasView::contentsMousePressEvent(pEvent);
		return;
	}
	
	// Take the zoom factor into consideration
	ptRealPos = pEvent->pos();
	ptRealPos /= m_dZoom;
	
	// Check if an item was clicked
	il = m_pCanvas->collisions(ptRealPos);
	for (itr = il.begin(); itr != il.end(); ++itr) {
		if (dynamic_cast<GraphNode*>(*itr) != NULL)
			pNode = dynamic_cast<GraphNode*>(*itr);
		else if (dynamic_cast<GraphEdge*>(*itr) != NULL)
			pEdge = dynamic_cast<GraphEdge*>(*itr);
	}
	
	// Handle clicks over different types of items
	if (pNode != NULL) {
		// Show a context menu for nodes
		showNodeMenu(pNode, pEvent->globalPos());
	}
	else if (pEdge != NULL) {
		// Show a context menu for edges
		showEdgeMenu(pEdge, pEvent->globalPos());
	}
	else {
		// Take the default action
		QCanvasView::contentsMousePressEvent(pEvent);
	}
}

/**
 * Sets defsault node attributes for the current graph.
 */
void GraphWidget::setNodeAttrib()
{
	Agnode_t* pNode;
	QFont font;
	
	font = Config().getFont(KScopeConfig::Graph);
	
	pNode = m_pGraph->proto->n;
	SET_NODE_ATTR(pNode, "shape", "box");
	SET_NODE_ATTR(pNode, "style", "filled");
	SET_NODE_ATTR(pNode, "fillcolor", 
		Config().getColor(KScopeConfig::GraphNode).name().latin1());
	SET_NODE_ATTR(pNode, "fontname", font.family().latin1());
	SET_NODE_ATTR(pNode, "fontsize",
		QString::number(font.pointSize()).latin1());
	SET_NODE_ATTR(pNode, "kscope_called", "f");
	SET_NODE_ATTR(pNode, "kscope_calling", "f");
}

/**
 * Changes the value of a node attribute which indicates whether a function
 * node has its called functions shown.
 * This is done by setting a custom node attribute.
 * @param	pNode	The node to set
 * @param	bShown	true if called functions are shown, false otherwise
 */
void GraphWidget::setCalledShown(GraphNode* pNode, bool bShown)
{
	SET_NODE_ATTR(pNode->getNode(), "kscope_called", 
		bShown ? ATTR_TRUE : ATTR_FALSE);	
}

/**
 * Changes the value of a node attribute which indicates whether a function
 * node has its calling functions shown.
 * @param	pNode	The node to set
 * @param	bShown	true if calling functions are shown, false otherwise
 */
void GraphWidget::setCallingShown(GraphNode* pNode, bool bShown)
{
	SET_NODE_ATTR(pNode->getNode(), "kscope_calling", 
		bShown ? ATTR_TRUE : ATTR_FALSE);	
}

/**
 * Determines whether called functions are shown for the given node.
 * This is done by reading the node attribute written by setCalledShown().
 * @param	pNode	The node to test
 */
bool GraphWidget::getCalledShown(GraphNode* pNode)
{
	return strcmp(GET_NODE_ATTR(pNode->getNode(), "kscope_called"), 
		ATTR_TRUE) == 0;
}

/**
 * Determines whether calling functions are shown for the given node.
 * This is done by reading the node attribute written by setCallingShown().
 * @param	pNode	The node to test
 */
bool GraphWidget::getCallingShown(GraphNode* pNode)
{
	return strcmp(GET_NODE_ATTR(pNode->getNode(), "kscope_calling"), 
		ATTR_TRUE) == 0;
}

/**
 * Removes all edges attached to a function node at the given direction.
 * Any strongly connected components that are no longer connected to that
 * function are deleted.
 * @param	pNode	The node for which to remove the edges
 * @param	bOut	true for outgoing edges, false for incoming
 */
void GraphWidget::removeEdges(GraphNode* pGraphNode, bool bOut)
{
	FirstEdgeFunc pFirst;
	NextEdgeFunc pNext;
	Agnode_t* pNode, * pDelNode, * pPrevNode;
	Agedge_t* pEdge;
	Agraph_t** pSubGraphs;
	int nSubGraphs, i;
	
	pNode = pGraphNode->getNode();
	
	// Set edge functions based on requested direction
	if (bOut) {
		pFirst = agfstout;
		pNext = agnxtout;
	}
	else {
		pFirst = agfstin;
		pNext = agnxtin;
	}
	
	// Nothing to do if no edges need to be removed
	if ((pEdge = pFirst(m_pGraph, pNode)) == NULL)
		return;
	
	// Remove all edges in the requested direction
	for (; pEdge != NULL; pEdge = pNext(m_pGraph, pEdge)) {
		if (bOut)
			agdelete(m_pGraph, pEdge->head);
		else
			agdelete(m_pGraph, pEdge->tail);
	}
	
	// Remove all subgraphs except the one that contains the given function
	pSubGraphs = ccomps(m_pGraph, &nSubGraphs, NULL);
	for (i = 0; i < nSubGraphs; i++) {
		if (!agcontains(pSubGraphs[i], pNode)) {
			// Remove all nodes in the subgraph
			for (pDelNode = agfstnode(pSubGraphs[i]); pDelNode != NULL;) {
				pPrevNode = pDelNode;
				pDelNode = agnxtnode(pSubGraphs[i], pDelNode);
				agdelete(m_pGraph, pPrevNode);
			}
			
			// Delete the subgraph
			agclose(pSubGraphs[i]);
		}
	}
}

/**
 * Shows a popup menu for a node.
 * This menu is shown after a node has been right-clicked.
 * @param	pNode	The node for which to show the menu
 * @param	ptPos	The position of the menu 
 */
void GraphWidget::showNodeMenu(GraphNode* pNode, const QPoint& ptPos)
{
	// Remember the node
	m_pMenuItem = pNode;
	
	// Show the popup menu.
	m_pNodePopup->popup(ptPos);
}

/**
 * Shows a popup menu for an edge.
 * This menu is shown after an edge has been right-clicked.
 * @param	pEdge	The edge for which to show the menu
 * @param	ptPos	The position of the menu 
 */
void GraphWidget::showEdgeMenu(GraphEdge* pEdge, const QPoint& ptPos)
{
	// Remember the edge
	m_pMenuItem = pEdge;
	
	// Show the popup menu.
	m_pEdgePopup->popup(ptPos);
}

/**
 * Shows functions called from the current function node.
 * This slot is connected to the "Show Called Functions" popup menu
 * action.
 */
void GraphWidget::slotShowCalled()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Request a query for called functions
	emit showCalled(pNode->getText());
}

/**
 * Shows functions calling tothe current function node.
 * This slot is connected to the "Show Calling Functions" popup menu
 * action.
 */
void GraphWidget::slotShowCalling()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Request a query for called functions
	emit showCalling(pNode->getText());
}

/**
 * Hides functions called from the current function node.
 * This slot is connected to the "Hide Called Functions" popup menu
 * action.
 */
void GraphWidget::slotHideCalled()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Remove edges and redraw the graph
	removeEdges(pNode, true);
	draw();
}

/**
 * Hides functions calling to the current function node.
 * This slot is connected to the "Hide CallingFunctions" popup menu
 * action.
 */
void GraphWidget::slotHideCalling()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	// Remove edges and redraw the graph
	removeEdges(pNode, false);
	draw();
}

/**
 * Looks up the definition of the current function node.
 * This slot is connected to the "Find Definition" popup menu action.
 */
void GraphWidget::slotFindDef()
{
	GraphNode* pNode;
	
	// Make sure the menu item is a node
	pNode = dynamic_cast<GraphNode*>(m_pMenuItem);
	if (pNode == NULL)
		return;
	
	emit findDef(pNode->getText());
}

/**
 * Emits a signal to open an editor at the file and line matching the call
 * information of the current edge.
 * This slot is connected to the "Open Call" popup menu action (for edges).
 */
void GraphWidget::slotOpenCall()
{
	GraphEdge* pEdge;
	QString sFile, sLine;
	
	// Make sure the menu item is an edge
	pEdge = dynamic_cast<GraphEdge*>(m_pMenuItem);
	if (pEdge == NULL)
		return;
	
	// Get call information from the edge
	sFile = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_file");
	sLine = GET_EDGE_ATTR(pEdge->getEdge(), "kscope_line");
	
	// Emit the signal
	emit lineRequested(sFile, sLine.toUInt());
}

#include "graphwidget.moc"
